/*
 * Copyright (C) 2000-2002 Red Hat, Inc.
 *
 * This is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Library General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include "../config.h"
#include <libintl.h>
#include <locale.h>
#include <popt.h>
#include <pwd.h>
#include <glade/glade-xml.h>
#include <gtk/gtk.h>
#include <krb5.h>

#ifndef KRB5_SUCCESS
#define KRB5_SUCCESS 0
#endif

#define _(String) dgettext(PACKAGE, String)

#define TOP "krb5"

static char *principal = NULL, *password = NULL, *realm = NULL,
	    *cachefile = NULL;
static GtkListStore *liststore = NULL;
static krb5_context context;
static krb5_ccache ccache;

static void
reread_creds(GladeXML *xml)
{
	krb5_cc_cursor cursor;
	krb5_creds creds;
	gchar **text = NULL;
	GtkTreeIter iter;
	char buf[LINE_MAX];
	int ret, row = 0;

	ret = krb5_cc_resolve(context, cachefile, &ccache);
	if (ret != KRB5_SUCCESS) {
		char *p = g_strdup_printf(_("Error resolving cache name: %s"),
					  error_message(ret));
		g_warning(p);
		return;
	}

	gtk_list_store_clear(liststore);
	ret = krb5_cc_start_seq_get(context, ccache, &cursor);
	if (ret == KRB5_SUCCESS) {
		while (krb5_cc_next_cred(context, ccache, &cursor, &creds) == KRB5_SUCCESS) {
			char *p = NULL, *q = NULL;
			krb5_timestamp starttime;
	
			text = g_malloc(sizeof(gchar*) * 3);
			krb5_unparse_name(context, creds.server, &text[2]);
			if (creds.times.starttime != 0) {
				starttime = creds.times.starttime;
			} else {
				starttime = creds.times.authtime;
			}
			text[0] = g_strdup_printf("%ld", starttime);
			if (krb5_timestamp_to_string(creds.times.starttime,
						    buf,
						    sizeof(buf)) == KRB5_SUCCESS) {
				g_free(text[0]);
				text[0] = g_strdup(buf);
			}
			text[1] = g_strdup_printf("%ld", creds.times.endtime);
			if (krb5_timestamp_to_string(creds.times.endtime,
						    buf,
						    sizeof(buf)) == KRB5_SUCCESS) {
				g_free(text[1]);
				text[1] = g_strdup(buf);
			}
			if (creds.ticket_flags != 0) {
				p = g_strdup_printf("%s (", text[2]);
				if (creds.ticket_flags & TKT_FLG_FORWARDABLE) {
					q = g_strdup_printf("%s%c", p, 'F');
					g_free(p);
					p = q;
				}
				if (creds.ticket_flags & TKT_FLG_FORWARDED) {
					q = g_strdup_printf("%s%c", p, 'f');
					g_free(p);
					p = q;
				}
				if (creds.ticket_flags & TKT_FLG_PROXIABLE) {
					q = g_strdup_printf("%s%c", p, 'P');
					g_free(p);
					p = q;
				}
				if (creds.ticket_flags & TKT_FLG_PROXY) {
					q = g_strdup_printf("%s%c", p, 'p');
					g_free(p);
					p = q;
				}
				if (creds.ticket_flags & TKT_FLG_MAY_POSTDATE) {
					q = g_strdup_printf("%s%c", p, 'D');
					g_free(p);
					p = q;
				}
				if (creds.ticket_flags & TKT_FLG_POSTDATED) {
					q = g_strdup_printf("%s%c", p, 'd');
					g_free(p);
					p = q;
				}
				if (creds.ticket_flags & TKT_FLG_INVALID) {
					q = g_strdup_printf("%s%c", p, 'X');
					g_free(p);
					p = q;
				}
				if (creds.ticket_flags & TKT_FLG_RENEWABLE) {
					q = g_strdup_printf("%s%c", p, 'R');
					g_free(p);
					p = q;
				}
				if (creds.ticket_flags & TKT_FLG_INITIAL) {
					q = g_strdup_printf("%s%c", p, 'I');
					g_free(p);
					p = q;
				}
				q = g_strdup_printf("%s%c", p, ')');
				g_free(p);
				p = q;
				text[2] = p;
			} else {
				p = g_strdup_printf("%s", text[0]);
				text[2] = p;
			}
			gtk_list_store_append(liststore, &iter);
			gtk_list_store_set(liststore, &iter,
					   0, text[0],
					   1, text[1],
					   2, text[2],
					   -1);
			g_free(text[0]);
			g_free(text[1]);
			g_free(text[2]);
		}
		krb5_cc_end_seq_get(context, ccache, &cursor);
		krb5_cc_close(context, ccache);
	}
}

static void
delete_creds(GtkWidget *widget, GladeXML *xml)
{
	GtkWidget *clist, *top, *dialog;
	int result;
	GError *error = NULL;

	top = glade_xml_get_widget(xml, TOP);
	g_assert(GTK_WIDGET_TOPLEVEL(top));
	dialog = gtk_message_dialog_new(GTK_WINDOW(top),
					GTK_DIALOG_MODAL |
					GTK_DIALOG_DESTROY_WITH_PARENT,
					GTK_MESSAGE_QUESTION,
					GTK_BUTTONS_YES_NO,
					_("Are you sure you want to delete your Kerberos credentials?"));
	result = gtk_dialog_run(GTK_DIALOG(dialog));
	gtk_widget_destroy(dialog);
	if (result == GTK_RESPONSE_YES) {
		if (g_spawn_command_line_sync(KDESTROY_PATH,
				 	      NULL,
					      NULL,
					      &result,
					      &error)) {
			reread_creds(xml);
		}
	}
}

static GValueArray *
get_realms(void)
{
	void *iterator;
	char *arealm;
	GValueArray *ret;
	GValue value;
	ret = g_value_array_new(1);
	memset(&value, 0, sizeof(value));
	if (krb5_realm_iterator_create(context, &iterator) == KRB5_SUCCESS) {
		while (krb5_realm_iterator(context, &iterator, &arealm) == KRB5_SUCCESS) {
			if (arealm != NULL) {
				g_value_init(&value, G_TYPE_STRING);
				g_value_set_string(&value, arealm);
				krb5_free_realm_string(context, arealm);
				g_value_array_append(ret, &value);
				g_value_unset(&value);
			}
		}
		krb5_realm_iterator_free(context, &iterator);
	}
	if (krb5_get_default_realm(context, &arealm) == KRB5_SUCCESS) {
		if (realm != NULL) {
			g_free(realm);
		}
		realm = g_strdup(arealm);
		krb5_free_default_realm(context, arealm);
	}
	return ret;
}

static void
punt_error(GtkWidget *top, const char *err)
{
	GtkWidget *dialog;
	dialog = gtk_message_dialog_new(GTK_WINDOW(top),
					GTK_DIALOG_MODAL |
					GTK_DIALOG_DESTROY_WITH_PARENT,
					GTK_MESSAGE_ERROR,
					GTK_BUTTONS_CLOSE,
					err);
	gtk_dialog_run(GTK_DIALOG(dialog));
	gtk_widget_destroy(dialog);
}

static void
get_creds(GtkWidget *widget, GladeXML *xml)
{
	GtkWidget *top, entry, *dialog;
	krb5_creds creds;
	int ret;
	char *p;

	top = glade_xml_get_widget(xml, TOP);
	if (!GTK_WIDGET_TOPLEVEL(top)) {
		top = NULL;
	}

	if ((principal == NULL) || (strlen(principal) == 0)) {
		return;
	}

	memset(&creds, 0, sizeof(creds));

	if ((strchr(principal, '@') == NULL) &&
	    (realm != NULL) &&
	    (strlen(realm) > 0)) {
		p = g_strdup_printf("%s@%s", principal, realm);
		ret = krb5_parse_name(context, p, &creds.client);
		g_free(p);
	} else {
		ret = krb5_parse_name(context, principal, &creds.client);
	}
	krb5_unparse_name(context, creds.client, &p);
	if (ret != KRB5_SUCCESS) {
		char *p = g_strdup_printf(_("Error parsing name to principal: %s"),
					  error_message(ret));
		punt_error(top, p);
		g_free(p);
		return;
	}

	ret = krb5_build_principal(context, &creds.server,
				   strlen(realm), realm,
				   KRB5_TGS_NAME, realm,
				   NULL);
	krb5_unparse_name(context, creds.server, &p);
	if (ret != KRB5_SUCCESS) {
		char *p = g_strdup_printf(_("Error parsing name to principal: %s"),
					  error_message(ret));
		punt_error(top, p);
		g_free(p);
		return;
	}

	if ((password == NULL) || (strlen(password) == 0)) {
		punt_error(top, _("No password entered!"));
		return;
	}

	ret = krb5_cc_resolve(context, cachefile, &ccache);
	if (ret != KRB5_SUCCESS) {
		char *p = g_strdup_printf(_("Error resolving cache name: %s"),
					  error_message(ret));
		punt_error(top, p);
		g_free(p);
		return;
	}

	ret = krb5_cc_initialize(context, ccache, creds.client);
	if (ret != KRB5_SUCCESS) {
		char *p;
		p = g_strdup_printf(_("Error creating credential cache: %s"),
				    error_message(ret));
		punt_error(top, p);
		g_free(p);
		return;
	}

	ret = krb5_get_in_tkt_with_password(context,
					    TKT_FLG_FORWARDABLE |
					    TKT_FLG_PROXIABLE |
					    TKT_FLG_RENEWABLE,
					    NULL,
					    NULL,
					    NULL,
					    password,
					    ccache,
					    &creds,
					    NULL);
	
	if (ret != KRB5_SUCCESS) {
		char *p;
		p = g_strdup_printf(_("Error getting credentials: %s"),
				    error_message(ret));
		punt_error(top, p);
		g_free(p);
	}
	krb5_cc_close(context, ccache);

	krb5_free_cred_contents(context, &creds);

	reread_creds(xml);
}

static void
update_string(GtkWidget *widget, gpointer data)
{
	g_assert(GTK_IS_ENTRY(widget));
	if (*(char**)data != NULL) {
		g_free(*(char**)data);
	}
	*((char**)data) = g_strdup(gtk_entry_get_text(GTK_ENTRY(widget)));
}


static void
clear_entry(GtkWidget *widget, GtkWidget *target)
{
	g_assert(GTK_IS_ENTRY(target));
	gtk_entry_set_text(GTK_ENTRY(target), "");
}

static void
change_password(GtkWidget *widget, GladeXML *xml)
{
	GtkWidget *entry, *toplevel, *dialog, *label, *table;
	int ret, i;
	krb5_creds creds;
	int result_code;
	char *pass = NULL, *npass1 = NULL, *npass2 = NULL;
	krb5_data result_string, result_data;

	memset(&creds, 0, sizeof(creds));
	memset(&result_string, 0, sizeof(result_string));
	memset(&result_data, 0, sizeof(result_data));

	/* Create a new dialog. */
	toplevel = glade_xml_get_widget(xml, TOP);
	dialog = gtk_message_dialog_new(GTK_WINDOW(toplevel),
					GTK_DIALOG_MODAL |
					GTK_DIALOG_DESTROY_WITH_PARENT,
					GTK_MESSAGE_QUESTION,
					GTK_BUTTONS_OK_CANCEL,
					_("Change Kerberos Password"));

	/* Create a table to store the prompts and answers. */
	table = gtk_table_new(3, 2, FALSE);
	gtk_box_pack_start_defaults(GTK_BOX((GTK_DIALOG(dialog))->vbox), table);
	gtk_widget_show(table);

	/* Create a new entry field for the old password. */
	label = gtk_label_new(_("Current password:"));
	gtk_table_attach_defaults(GTK_TABLE(table), label, 0, 1, 0, 1);
	gtk_widget_show(label);
	entry = gtk_entry_new();
	gtk_entry_set_visibility(GTK_ENTRY(entry), FALSE);
	if (password != NULL) {
		gtk_entry_set_text(GTK_ENTRY(entry), password);
	}
	gtk_signal_connect(GTK_OBJECT(entry), "changed",
			   GTK_SIGNAL_FUNC(update_string), &pass);
	gtk_table_attach_defaults(GTK_TABLE(table), entry, 1, 2, 0, 1);
	gtk_widget_show(entry);

	/* Create an entry for the new password. */
	label = gtk_label_new(_("New password"));
	gtk_table_attach_defaults(GTK_TABLE(table), label, 0, 1, 1, 2);
	gtk_widget_show(label);

	entry = gtk_entry_new();
	gtk_entry_set_visibility(GTK_ENTRY(entry), FALSE);
	gtk_signal_connect(GTK_OBJECT(entry), "changed",
			   GTK_SIGNAL_FUNC(update_string), &npass1);
	gtk_table_attach_defaults(GTK_TABLE(table), entry, 1, 2, 1, 2);
	gtk_widget_show(entry);

	/* Create an entry to verify the new password. */
	label = gtk_label_new(_("New password (verify)"));
	gtk_table_attach_defaults(GTK_TABLE(table), label, 0, 1, 2, 3);
	gtk_widget_show(label);

	entry = gtk_entry_new();
	gtk_entry_set_visibility(GTK_ENTRY(entry), FALSE);
	gtk_signal_connect(GTK_OBJECT(entry), "changed",
			   GTK_SIGNAL_FUNC(update_string), &npass2);
	gtk_table_attach_defaults(GTK_TABLE(table), entry, 1, 2, 2, 3);
	gtk_widget_show(entry);

	do {
		ret = gtk_dialog_run(GTK_DIALOG(dialog));
		if (ret == GTK_RESPONSE_CANCEL) {
			gtk_widget_destroy(dialog);
			return;
		}
		if (strcmp(npass1, npass2) != 0) {
			punt_error(toplevel, _("New passwords do not match"));
			continue;
		}
		if (ret == GTK_RESPONSE_OK) {
			break;
		}
	} while (1);

	/* Convert the user's name into a principal. */
	ret = krb5_parse_name(context, principal, &creds.client);
	if (ret != KRB5_SUCCESS) {
		char *p;
		gtk_widget_destroy(dialog);
		p = g_strdup_printf(_("Error parsing principal \"%s\": %s"),
				    principal, error_message(ret));
		punt_error(toplevel, p);
		g_free(p);
		return;
	}

	/* Convert the changepw service name into a principal. */
	ret = krb5_parse_name(context, "kadmin/changepw", &creds.server);
	if (ret != KRB5_SUCCESS) {
		char *p;
		gtk_widget_destroy(dialog);
		p = g_strdup_printf(_("Error parsing principal \"%s\": %s"),
				    principal, error_message(ret));
		punt_error(toplevel, p);
		g_free(p);
		return;
	}

	/* Get a password-change credential. */
	ret = krb5_get_in_tkt_with_password(context,
					    0,
					    NULL,
					    NULL,
					    NULL,
					    pass,
					    NULL,
					    &creds,
					    NULL);
	if (ret != KRB5_SUCCESS) {
		char *p;
		gtk_widget_destroy(dialog);
		p = g_strdup_printf(_("Error obtaining password-changing credentials: %s"),
				    error_message(ret));
		punt_error(toplevel, p);
		g_free(p);
		return;
	}

	/* Try to perform the change. */
	ret = krb5_change_password(context, &creds, npass1, &result_code,
				   &result_string, &result_data);
	if (ret != KRB5_SUCCESS) {
		char *p;
		gtk_widget_destroy(dialog);
		p = g_strdup_printf(_("Error changing password for \"%s\": %s"),
				    principal, error_message(ret));
		punt_error(toplevel, p);
		g_free(p);
	}

	/* Clean up. */
	gtk_widget_destroy(dialog);
	krb5_free_cred_contents(context, &creds);
}

int
main(int argc, char **argv)
{
	GValueArray *realms = NULL;
	GValue *value;
	GladeXML *xml;
	GtkWidget *widget, *entry;
	GtkTreeViewColumn *column;
	GtkCellRenderer *renderer;
	GList *list;
	int ret, i;
	krb5_principal princ;

	/* Initialize the locale settings. */
	setlocale(LC_ALL, "");
	bindtextdomain(PACKAGE, GNOMELOCALEDIR);
	textdomain(PACKAGE);

	/* Initialize Kerberos. */
	initialize_krb5_error_table();
	if ((ret = krb5_init_context(&context)) != KRB5_SUCCESS) {
		g_error("Error initializing Kerberos: %s", error_message(ret));
		return 1;
	}

	/* Initialize GTK+. */
	gtk_init(&argc, &argv);

	/* Load the GUI. */
	xml = glade_xml_new(DATADIR "/" PACKAGE "/" PACKAGE ".glade",
			    TOP, PACKAGE);
	if (xml == NULL) {
		xml = glade_xml_new(PACKAGE ".glade", TOP, PACKAGE);
	}
	if (xml == NULL) {
		g_error("Error initializing libglade!");
		return 1;
	}

	/* Set the default ccache name. */
	cachefile = g_strdup(krb5_cc_default_name(context));
	ret = krb5_cc_set_default_name(context, cachefile);
	if (ret != KRB5_SUCCESS) {
		g_error("Error setting default cache name: %s",
			error_message(ret));
		return 1;
	}
	widget = glade_xml_get_widget(xml, "filename");
	if (GTK_IS_LABEL(widget)) {
		gtk_label_set_text(GTK_LABEL(widget), cachefile);
	}

	/* Resolve the cache to a ccache structure.  This mainly maps the
	 * ccache type to a set of methods for manipulating it. */
	ret = krb5_cc_resolve(context, cachefile, &ccache);
	if (ret != KRB5_SUCCESS) {
		g_error("Error resolving cache name: %s", error_message(ret));
		return 1;
	}

	/* Get the default principal's name and set it in the right label. */
	ret = krb5_cc_get_principal(context, ccache, &princ);
	if (ret == KRB5_SUCCESS) {
		char *tmp;
		ret = krb5_unparse_name(context, princ, &principal);
		if (principal != NULL) {
			tmp = g_strdup(principal);
			krb5_free_unparsed_name(context, principal);
			principal = tmp;
		}
	} else {
		struct passwd *pwd;
		pwd = getpwuid(getuid());
		if (pwd != NULL) {
			principal = g_strdup(pwd->pw_name);
		}
	}
	krb5_cc_close(context, ccache);

	/* Get the list of configured realms and set them up in the tool's
	 * pop-down realms list. */
	realms = get_realms();
	widget = glade_xml_get_widget(xml, "realmcombo");
	if (realms->n_values > 0) {
		list = NULL;
		for (i = 0; i < realms->n_values; i++) {
			value = g_value_array_get_nth(realms, i);
			list = g_list_append(list,
					     (char*)g_value_get_string(value));
		}
		gtk_combo_set_popdown_strings(GTK_COMBO(widget), list);
		if (realm != NULL) {
			widget = (GTK_COMBO(widget))->entry;
			if (GTK_IS_ENTRY(widget)) {
				gtk_entry_set_text(GTK_ENTRY(widget), realm);
			}
		}
		if (list != NULL) {
			g_list_free(list);
		}
		g_value_array_free(realms);
	}

	/* Set up the principal and password entry fields to update the
	 * related strings. */
	widget = glade_xml_get_widget(xml, "principalentry");
	if (GTK_IS_ENTRY(widget)) {
		gtk_entry_set_text(GTK_ENTRY(widget), principal);
		g_signal_connect(G_OBJECT(widget), "changed",
				 G_CALLBACK(update_string), &principal);
	}
	widget = glade_xml_get_widget(xml, "passwordentry");
	if (GTK_IS_ENTRY(widget)) {
		g_signal_connect(G_OBJECT(widget), "changed",
				 G_CALLBACK(update_string), &password);
	}
	widget = glade_xml_get_widget(xml, "realmentry");
	if (GTK_IS_ENTRY(widget)) {
		g_signal_connect(G_OBJECT(widget), "changed",
				 G_CALLBACK(update_string), &realm);
	}

	/* Set up the list store for use by the TreeView in the dialog. */
	widget = glade_xml_get_widget(xml, "credentials");
	if (GTK_IS_TREE_VIEW(widget)) {
		liststore = gtk_list_store_new(3,
					       G_TYPE_STRING,
					       G_TYPE_STRING,
					       G_TYPE_STRING);
		gtk_tree_view_set_model(GTK_TREE_VIEW(widget),
					GTK_TREE_MODEL(liststore));

		column = gtk_tree_view_column_new();
		renderer = gtk_cell_renderer_text_new();
		gtk_tree_view_column_pack_start(column, renderer, TRUE);
		gtk_tree_view_column_set_title(column, _("Valid Starting"));
		gtk_tree_view_column_set_attributes(column, renderer,
						    "text", 0, NULL);
		gtk_tree_view_insert_column(GTK_TREE_VIEW(widget), column, 0);

		column = gtk_tree_view_column_new();
		renderer = gtk_cell_renderer_text_new();
		gtk_tree_view_column_pack_start(column, renderer, TRUE);
		gtk_tree_view_column_set_title(column, _("Expires At"));
		gtk_tree_view_column_set_attributes(column, renderer,
						    "text", 1, NULL);
		gtk_tree_view_insert_column(GTK_TREE_VIEW(widget), column, 1);

		column = gtk_tree_view_column_new();
		renderer = gtk_cell_renderer_text_new();
		gtk_tree_view_column_pack_start(column, renderer, TRUE);
		gtk_tree_view_column_set_title(column, _("Principal"));
		gtk_tree_view_column_set_attributes(column, renderer,
						    "text", 2, NULL);
		gtk_tree_view_insert_column(GTK_TREE_VIEW(widget), column, 2);
	}

	/* Hook up callbacks for the buttons. */
	widget = glade_xml_get_widget(xml, "delete");
	if (GTK_IS_BUTTON(widget)) {
		g_signal_connect(G_OBJECT(widget), "clicked",
			       	 G_CALLBACK(delete_creds), xml);
	}
	widget = glade_xml_get_widget(xml, "passwd");
	if (GTK_IS_BUTTON(widget)) {
		g_signal_connect(G_OBJECT(widget), "clicked",
			       	 G_CALLBACK(change_password), xml);
	}
	widget = glade_xml_get_widget(xml, "getcred");
	if (GTK_IS_BUTTON(widget)) {
		g_signal_connect(G_OBJECT(widget), "clicked",
			       	 G_CALLBACK(get_creds), xml);
		/* Add an extra callback to clear the password-entry field. */
		entry = glade_xml_get_widget(xml, "passwordentry");
		if (GTK_IS_ENTRY(entry)) {
			g_signal_connect(G_OBJECT(widget), "clicked",
				       	 G_CALLBACK(clear_entry), entry);
		}
	}

	/* Refresh the creds list. */
	reread_creds(xml);

	/* Show the main window. */
	widget = glade_xml_get_widget(xml, TOP);
	g_signal_connect(G_OBJECT(widget), "delete_event",
			 G_CALLBACK(gtk_main_quit), NULL);
	gtk_widget_show_all(widget);
	gtk_main();

	return 0;
}
